# $Id: fox.gap,v 1.11 2002/08/15 01:30:56 nathand Exp $
#
# The main functions contained here are 
# 
#   RankOfHomologyWithLocalCoeffs, PresentHomologyWithLocalCoeffs 
#
# which compute the 1st homology of a group G with coeffs in a module
# V, which is a vector space over some field K.  The functions take in
# (G, ImagesOfGens) where ImagesOfGens are the matrices (with entries
# in K) corresponding to the action of the generators of G on the
# module V.
#
# There is also the function HomologyOfRegularCoverViaLocalCoeffs
# which given a group G and a homomorphism h computes the 1st
# Q-homology of the kernel of h using the above functions.  

#  THE FUNCTIONS: FoxGen, ForDer, MapGroupRing, OldPresentHomology
# with local coeffs are no longer used.  

# Computes  d g2 / d g1 in free group F into group ring R

FoxGen := function(g1, g2, F, R, emb)
  if g1 = g2 then 
    return Identity(F)^emb;
  fi;
  if g1 = g2^-1 then 
    return -g2^emb;
  fi;
  return Zero(R);
end;


FoxDerivative := function(var, word, F, R, emb)
    if IsOne(word) then return Zero(R); fi;
  return Sum(List([1..Length(word)],
    k -> Subword(word,1,k-1)^emb * FoxGen(var, Subword(word,k,k), F, R, emb) ));
end;

# helper function

MakeBlockMatrix := function( blocks )
  return Concatenation( 
    List(blocks, rowofblocks -> 
       TransposedMat(Concatenation(
           List(rowofblocks, block -> TransposedMat(block) )
     )))
   );
end;

# The function MapGroupRing takes the inverse of Image(f, C[k])
# because the right module sturcture on ZG when forming the tensor
# product is _not_ the natural one, but the one induced from the
# natural left module structure.  I.e., m.g = g^-1.m.  This means when
# flipping everything over to the other side, we have to take an
# inverse.  See, Brown, Cohomology of Groups, Chap III.

MapGroupRing := function(r, f)
   local C;
   C := CoefficientsAndMagmaElements(r);
   return Sum(Filtered([1..Length(C)], x -> x mod 2 = 1), 
      k -> C[k+1] * Inverse(Image(f, C[k])) );
end;
  

# Computes a presentation of the homology of G with coefficents in the
# K-vector space V.  Takes a finitely presented group G, list of
# K-matrices which record the action of the generators of G on V (in
# terms of the same basis).
#
# Returns the pair of matrixes A,
# B which present the homology of the kernel of h as Ker(A)/Image(B).


OldPresentHomologyWithLocalCoeffs := function(G, ImagesOfGen)
   local F, R, emb, M, g, r, f;
   F := FreeGroupOfFpGroup(G);
   R := GroupRing(Integers, F);
   emb := Embedding(F, R);
   M := Group(ImagesOfGen);
   f := GroupHomomorphismByImagesNC(F, M, GeneratorsOfGroup(F), ImagesOfGen);
   return [
    
      MakeBlockMatrix(
       [List(GeneratorsOfGroup(F), g -> MapGroupRing(g^emb - One(R), f))])
     ,
    MakeBlockMatrix(
     List(GeneratorsOfGroup(F), g ->
       List(RelatorsOfFpGroup(G), r -> 
           MapGroupRing(FoxDerivative(g, r, F, R, emb), f)
       )
     )
   )];
end;


# New improved version should be much faster.

MatrixOfMappedFoxDerivative := function(var, word, n, F, imofgen, inverseimofgen)
    local D, currprod, currlet, invword, i;
    D := NullMat(n,n);
    currprod := IdentityMat(n,n);
    invword := word^-1;
    for i in Reversed([1..Length(word)]) do
      currlet := Subword(invword,i,i);
      if currlet = var then 
         currprod :=  imofgen[GeneratorSyllable(currlet, 1)] * currprod;
         D := D - currprod;
      else
        if currlet = var^-1 then
          D := D + currprod;
        fi;
        if ExponentSyllable(currlet, 1) > 0 then 
           currprod := imofgen[GeneratorSyllable(currlet, 1)] * currprod;
        else
           currprod := inverseimofgen[GeneratorSyllable(currlet, 1)] * currprod;
        fi;
      fi;
    od;
    return D;
end;

 
BlocksForHomologyWithLocalCoeffs := function(G, ImagesOfGen)
   local F, g,r,n,InverseImagesOfGen;
   InverseImagesOfGen := List(ImagesOfGen, x -> Inverse(x));
   F := FreeGroupOfFpGroup(G);
   n := DimensionsMat(ImagesOfGen[1])[1];
   return [
       InverseImagesOfGen,
  
     List(GeneratorsOfGroup(F), g ->
       List(RelatorsOfFpGroup(G), r -> 
           MatrixOfMappedFoxDerivative(g, r, n, F,
       ImagesOfGen, InverseImagesOfGen)
      )
   )
   ];
end;

PresentHomologyWithLocalCoeffsFromBlocks := function( ABlocks, BBlocks )
   local n;
   n := DimensionsMat(ABlocks[1])[1];
   return [ 
    MakeBlockMatrix(
     [List(ABlocks, A -> 
           A - IdentityMat(n,n)  )] )
     ,
     MakeBlockMatrix( BBlocks )];
end;

PresentHomologyWithLocalCoeffs := function(G, ImagesOfGen)
  local tmp;
  tmp := BlocksForHomologyWithLocalCoeffs(G, ImagesOfGen);
  return PresentHomologyWithLocalCoeffsFromBlocks(tmp[1], tmp[2]);
end;

RankOfHomologyWithLocalCoeffs := function(G, ImagesOfGen)
   local M, A, B;
   M := PresentHomologyWithLocalCoeffs(G, ImagesOfGen);
   A := M[1];
   B := M[2];
   if not IsZero(A*B) then
     Error("Boundary^2 is not 0!");
   fi;
   return (DimensionsMat(A)[2] - RankMat(A)) - RankMat(B);
   #return [ A, B, DimensionsMat(A),  RankMat(A), DimensionsMat(B), RankMat(B)];
end;

# Entries must be integers!

QuickRankOfHomologyWithLocalCoeffs := function(G, ImagesOfGen)
   local M, A, B;
   M := PresentHomologyWithLocalCoeffs(G, ImagesOfGen);
   A := M[1];
   B := M[2];
   if not IsZero(A*B) then
     Error("Boundary^2 is not 0!");
   fi;
   
   return (DimensionsMat(A)[2] 
          - RankMatModP(A, 31991)) - RankMatModP(B, 31991);
end;

QuickRankOfHomologyWithFromBlocks := function(ABlocks, BBlocks)
   local M, A, B;
   M := PresentHomologyWithLocalCoeffsFromBlocks(ABlocks, BBlocks);
   A := M[1];
   B := M[2];
   if not IsZero(A*B) then
     Error("Boundary^2 is not 0!");
   fi;
   return (DimensionsMat(A)[2] 
          - RankMatModP(A, 31991)) - RankMatModP(B, 31991);
end;


HomologyOfRegularCoverViaLocalCoeffs := function(G, h)
  local Q, ImagesOfGen, f, g;
  Q := Image(h);
  f := ActionHomomorphism(Q, Q, OnRight); 
  ImagesOfGen := List(GeneratorsOfGroup(G),  
        g ->  PermutationMat(Image(f, Image(h, g)), Order(Q) )  );
  return RankOfHomologyWithLocalCoeffs(G, ImagesOfGen);
end;
    

TestHomologyWithLocalCoeffs := function(manifold)
        local G, f, maps, homology, rank1, rank2;
        G := manifold.Group();
        maps := GQuotients(G, PSL(2,5));
        for f in maps do
                Print("Cover found for ", manifold.Name, ": betti = [");
                rank1 := AbelianRankOfSubgroup(G, Kernel(f));
                rank2 := HomologyOfRegularCoverViaLocalCoeffs(G, f);
                Print(rank1, " , ", rank2, " ]\n");
        od;
        return false;
end;



   
    
    
    
                   
                  