# $Id: mfld_base.gap,v 1.23 2002/01/25 22:06:46 nathand Exp $
#
# Manifold Class implemented as a record (note format revised 2001/3/10):
#
#  M := rec(
#		Name := "SnapPea name";
#               Volume := "3.141592653";
#		Homology := [0,0,3,6];  
#		Group := Function which when called returns the group;
#		KnownPosBettiCover := false;
#               KnownWeakPosBettiCover := false;   # means mod a large prime
#		GoodCover := false; 
#                 # If Known := true, should be a tuple
#                 # [subgroup type, subgroup identifier, additional data]
#                 # Acceptable types are:
#                 #  ["Abelian"]
#                 #  ["LowIndex", index]
#                 #  ["PSL2", p]
#                 #  ["Perfect", ident, name]
#                 #  ["p-group", [prime, height]]
#                 #  ["IntersectionOfLowIndex, [index1, index2]]
#               GoodCoverImage := [(1,2,3), (4,5)],
#                 #  A list of permutations so that sending the 
#                 #  generators of Group() to these permutations
#                 #  induces a homorphism whose kernel has
#                 #  positive betti number
#               GoodCoverImageU := [(1,2,3)]
#                 #  A subgroup of GoodCoverImage so that if f is the 
#                 #  homomorphism from M.Group() to GoodCoverImage then
#                 #  f^(-1)(GoodCoverImageU) has positive betti number
#                 #
#                 #  This field is usually only included when 
#                 #  GoodCoverImage is so huge that reconstructing 
#                 #  GoodCoverImageU would take a long time.
#		Reason := false,
#                 #  Which program was used to check the homology
#               Degree :=0,   #  the degree of the cover which has pos. betti.
#                             #  num (use same degree as cover used for Rank
#               Rank := 0     #  Rank of homology of pos. betti. cover.
#                 
#
#          );
#
#  Other fields added as desired, on the fly, but saving is problematic.
#  It seems that saving self-contained functions usually works.
#
# Manifolds are assumed to be stored together in a list, in the directory 
# "manifolds/".

# Returns the list of manifolds contained in file manifoldfile
# All files end in ".gap"

LoadManifolds := function(manifoldfile)
	local f;
	f := ReadAsFunction(Concatenation("manifolds/", manifoldfile));
	return f();
end;

SaveManifolds := function(manifoldfile, manifolds)
	PrintTo(Concatenation("manifolds/", manifoldfile), "return ", manifolds, ";\n");
end;

# Returns the manifold with of SnapPea name "name" from manifolds

GetManifold := function(manifolds, name)
	local manifold;
	for manifold in manifolds do
		if manifold.Name = name then
			return manifold;
		fi;
	od;
	return fail;
end;

# Functions for selecting manifolds satisfying certain criteria 
# for the list manifolds

ManifoldsWithoutPosBettiCovers := function(manifolds)
	return Filtered(manifolds, x -> not x.KnownPosBettiCover);
end;

ManifoldsWithPosBettiCovers := function(manifolds)
	return Filtered(manifolds, x -> x.KnownPosBettiCover);
end;

ManifoldsWithoutWeakPosBettiCovers := function(manifolds)
	return Filtered(manifolds, x -> not x.KnownWeakPosBettiCover);
end;

ManifoldsWithWeakPosBettiCovers := function(manifolds)
	return Filtered(manifolds, x -> x.KnownWeakPosBettiCover);
end;

ManifoldAlmostDone := function(manifold)
    return manifold.KnownPosBettiCover or IsBound(manifold.GoodCoverImageU);
end;


ManifoldsNotAlmostDone := function(manifolds)
	return Filtered(manifolds, x -> not (x.KnownPosBettiCover or IsBound(x.GoodCoverImageU)));
end;

ManifoldsAlmostDone := function(manifolds)
	return Filtered(manifolds, x -> x.KnownPosBettiCover or IsBound(x.GoodCoverImageU));
end;


# If a cover with pos. 1st. betti number is found, use this
# function to record in the census.  DEPRECATED, not used in rigour.gap

FoundPosBettiCover := function(manifold, cover, reason)
	manifold.KnownPosBettiCover := true;
	manifold.GoodCover := cover;
	manifold.Reason := reason;
end;
		
ManifoldStats := function(manifolds)
	Print( Length(manifolds), " manifolds of which ", Number(manifolds, x -> x.KnownPosBettiCover), " have covers with pos.betti.num.\n");
end;


# Below functions added for saving just the info about covers, etc,
# without the actual group inorder to save space.

SaveCondensedManifolds := function(manifoldfile, manifolds)
	local m,C;
        C := StructuralCopy(manifolds);
        for m in C do
                Unbind(m.Group);
                Unbind(m.Homology);
                Unbind(m.Volume);
        od;
        PrintTo(Concatenation("manifolds/", manifoldfile), "return ", C, ";\n");
end;


UpdateManifoldsFromFile := function(manifoldfile, manifolds)
        local othermanifolds, M, Mname, N, updatenames, mm;
        othermanifolds := LoadManifolds(manifoldfile);
        updatenames := Intersection(
                List(ManifoldsWithoutPosBettiCovers(manifolds),mm->mm.Name),
                List(ManifoldsWithPosBettiCovers(othermanifolds),mm->mm.Name));
        for Mname in updatenames do
            M := GetManifold(manifolds, Mname);
            N := GetManifold(othermanifolds, Mname);
            FoundPosBettiCover(M, N.GoodCover, N.Reason);
         od;
end;


OverwriteManifoldsFromFile := function(manifoldfile, manifolds)
        local othermanifolds, M, Mname, N, updatenames, mm;
        othermanifolds := LoadManifolds(manifoldfile);
        for M in manifolds do
            N := GetManifold(othermanifolds, M.Name);
            if N.KnownPosBettiCover then
              FoundPosBettiCover(M, N.GoodCover, N.Reason);
            fi;
         od;
end;

# takes any improved data in manifolds2 that is 
# not in manifolds1 and puts it into manifolds1
# returns nothing.

MergeManifolds := function(manifolds1, manifolds2)
    local M, N, i;
    for i in [1..Length(manifolds1)] do
        M := manifolds1[i];
        if not M.KnownPosBettiCover then 
            N := GetManifold(manifolds2, M.Name);
            if N <> fail then
                if N.KnownPosBettiCover or
                   (Length(M.GoodCoverImage) = 0 and Length(N.GoodCoverImage) > 0) or
                   ( (not IsBound(M.Comment)) and IsBound(N.Comment) )
                   then
                    manifolds1[i] := N;
                    Print(N.Name, "\n");
                fi;
            fi;
        fi;
    od;
end;

    
StoreGroupHom := function(G, f)
   return Group( List(GeneratorsOfGroup(G), function(g) return Image(f,g); end));
end;

RandomManifolds := function(manifolds, n)
   local i, L, M;
   L := [];
   for i in [1..n] do
     repeat 
        M := Random(manifolds);
     until Number(L, x -> x = M) = 0;
     Add(L, M);
   od;
   return L;
end;

GoodCoverIsPSLCover := function(M)
   if M.KnownPosBettiCover then
      if M.GoodCover[1] = "PSL" or M.GoodCover[1] = "PSL2" then
         return true;
      fi;
   fi;
   return false;
end;

Has2GeneratorGroup := function(M)
   return Length(GeneratorsOfGroup(M.Group())) = 2;
end;

LoadManifoldsAlt := function(manifoldfile)
	local f;
	f := ReadAsFunction(Concatenation("/home/nathand/home/thurston/manifolds/", manifoldfile));
	return f();
end;

ClearCoverData := function(manifold)
    manifold.KnownPosBettiCover := false;
    manifold.KnownWeakPosBettiCover := false;
    manifold.Degree := 0;
    manifold.GoodCover := false;
    manifold.Reason := false;
    manifold.Rank := 0;
    manifold.GoodCoverImage := [];
    Unbind(manifold.GoodCoverImageU);
end;


# CreateManifoldFromGroup
#
# Function to create a "manifold" from a group
# The first function is a helper fns.


CreateGroupFunctionFromGroup := function(group)
    local group_str, group_stream, snipit, g, S, i, fun_group;
    
    # First, create a function that recreates the given group.  
    # This is a lot harder that it should be because GAP can't save
    # a finitely presented group.  
    
    group_str := "";
    group_stream := OutputTextString(group_str, true);
    S :=  FreeGeneratorsOfFpGroup(group);
    snipit := String(S);
    PrintTo(group_stream, "local F, ",  snipit{[2..(Length(snipit) - 1)]},
            ";\n"  );
    PrintTo(group_stream, "F := FreeGroup(", List(S, String), ");\n");
    for i in [1..Length(S)] do
        PrintTo(group_stream, S[i], " := F.", i, ";\n");
    od;
    PrintTo(group_stream, "return F/ ", RelatorsOfFpGroup(group), ";\n");
    CloseStream(group_stream);
    group_stream := InputTextString(group_str);
    fun_group := ReadAsFunction(group_stream);
    CloseStream(group_stream);
    return fun_group;
end;

CreateManifoldFromGroup := function(name,group)
    local M, G, group_fn, H;
    group_fn := CreateGroupFunctionFromGroup(group);
    G := group_fn();
    H := AbelianInvariants(G);
    M := rec(Name := name, Homology := H, Group := group_fn,
             KnownPosBettiCover := false, KnownWeakPosBettiCover := false,
             GoodCover := false, GoodCoverImage := [],
             Reason := false, Degree:= 0, Rank := 0);
    return M;
end;


         
    
    
    
    
    



Read("quick_rank.gap"); Read("fox.gap"); Read("group_lib.gap"); Read("search.gap"); Read("rigour.gap"); Read("other_groups.gap");

