# $Id: other_groups.gap,v 1.4 2002/06/07 21:09:32 nathand Exp $
#
# Tools for analyzing groups and making groups with similar statitisics,
# for purpose of compareson.


#-------------------------------------------------------------------------
#
# Code for analyzing the statistics of the presentation
# 
#-------------------------------------------------------------------------

# Returns the transition matrix T where 
# T[i][j] is the prob of transfering from state 
# i to state j.  

RelatorTransitions := function(rel, S)
    local T, curr, next, i, n;
    T := NullMat(Length(S), Length(S));
    n := Length(rel);
    for i in [1..n] do
        curr := Position(S, Subword(rel,i, i));
        if i <> n then 
            next := Position(S, Subword(rel,i + 1,i+1));
        else
            next := Position(S, Subword(rel,1, 1));
        fi;
        T[curr][next] := T[curr][next] + 1;
    od;
    return T;
end;

TestTransitionStatistics := function(G)
    local gens, rels, S, Trans, g;
    
    # Make symmetric generating set
    gens := FreeGeneratorsOfFpGroup(G);
    S := [];
    for g in gens do
        Add(S, g);
        Add(S, g^-1);
    od;
    
    rels := RelatorsOfFpGroup(G);
    
    Trans := List(rels, rel ->  RelatorTransitions(rel, S));
    return Trans;
end;

TestOverallStats := function(manifolds)
    local T, t, M;
    T := NullMat(4,4);
    for M in manifolds do
        t := TestTransitionStatistics(M.Group());
        T := T + t[1] + t[2];
    od;
    return T;
end;

PrettyPrintTransStats := function(stats_mat)
    local row, new_row, r;
    for row in stats_mat do
        new_row := row/Sum(row);
        for r in new_row do
            Print(FixedLengthThreeDigitDecimal(r), "  ");
        od;
        Print("\n");
    od;
end;    

# Surprisingly(?) for the simple1000 manifolds the distribution is far
# from uniform.  Here it is (where the [i,j] entry is the prob.of
# going from state i to state j).
#
#       a       A        b       B
#  a  0.435   0.000   0.406   0.157  
#  A  0.000   0.271   0.505   0.222  
#  b  0.386   0.140   0.473   0.000  
#  B  0.505   0.209   0.000   0.285 
#
#  Overall dist. of the various letters.
#     [  a   ,  A  ,   b  ,  B   ]
#     [ 20564, 6008, 21646, 6415 ]
#
#  [ [ 8950, 0, 8365, 3249 ],
#    [ 0, 1632, 3039, 1337 ], 
#    [ 8372, 3032, 10242, 0 ], 
#    [ 3242, 1344, 0, 1829 ] ]

# Berge observed the following for two generator presentations
# arising from Heegaard splittings (it's actually obvious)
#
# Namely, for the generator "a" there are intergers s,t >= 0 so
# that if a^e is a syllable of a relator then e = +/-s, t, or (s + t)
#
# The function "hasBergeForm" takes a two generator presentation and 
# checks if it has this form and returns [ has_form, [s_a, t_a], [ s_b, t_b]]
# 

exponentsOfGeneratorInRelation := function(R, genNum)
    return List(Filtered([1..NumberSyllables(R)],
                   i -> GeneratorSyllable(R, i) = genNum), 
                j -> AbsInt(ExponentSyllable(R, j)));
end;

exponentsOfGenerator := function(G, genNum)
    local L;
    L := Unique(Concatenation(List(RelatorsOfFpGroup(G), R -> exponentsOfGeneratorInRelation(R, genNum))));
    Sort(L);
    return L;
end;

generatorHasBergeForm := function(G, genNum)
    local E;
    E := exponentsOfGenerator(G, genNum);
    if Length(E) <= 1 then return true; fi;
    return (Length(E) <= 2 or ( Length(E) = 3 and E[1] + E[2] = E[3])) and GcdInt(E[1], E[2]) = 1;
end;
    
hasBergeForm := function(G)
    local berge;
    berge := not (false in List([1..Length(GeneratorsOfGroup(G))],
                     i -> generatorHasBergeForm(G, i)));
    return berge;
end;

# Code for creating new groups based on berge's picture

RandomRelatorFromBergeForm := function(G, possExponents, relatorToCopy)
    local n, E, F, i, word, newexp;
    F := FreeGroupOfFpGroup(G);
    n := Length(GeneratorsOfGroup(G));
    i := Random([1..n]);
    word := One(F);
    while Length(word) < Length(relatorToCopy) do
        newexp :=Random(possExponents[i])*Random([-1,1]);
        word := word * (GeneratorsOfGroup(F)[i])^(newexp);
        i := Random( Concatenation( [1..i-1], [i+1..n] ));
    od;
    return word;
end;

MakeGroupWithBergeForm := function(G)
    local n, possExponents;
    if not hasBergeForm(G) then
        return TrivialGroup();
    fi;
    n := Length(GeneratorsOfGroup(G));
    possExponents := List([1..n], i -> exponentsOfGenerator(G, i));
    return FreeGroupOfFpGroup(G)/List(RelatorsOfFpGroup(G), R ->
                   RandomRelatorFromBergeForm(G, possExponents, R));
end;

    
       
        
    

#---------------------------------------------------------------------
#
# Code for creating presentation whose transistion statistics are 
# those of the given group
#
#---------------------------------------------------------------------

# Here D is something like [2,4,5,1]

RandomAccordingToDistribution := function(S, D)
    local poss;
    if Maximum(D) = 0 then
        D := List([1..Length(S)], x -> 1);
    fi;
    
    poss := Concatenation(
                    List([1..Length(S)], i -> List([1..D[i]], x -> S[i])));
    return Random(poss);
end;

# S = generating set, T = transition matrix

CreateWordAccordingToTransitions:= function(S, T)
    local w, len, i;
    len := Sum(Sum(T));
    
    # Pick first letter, with dist. proportional to the orig.
    # word.
    
    w := RandomAccordingToDistribution(S, Sum(T));
    for i in [1..(len - 1)] do
        w := w * RandomAccordingToDistribution(S, 
                     T[Position(S, Subword(w, i, i))]);
    od;
    return w;
end;


# When generating a reduced word, have to worry about that
# last element.

CreateReducedWordAccordingToTransitions:= function(S, T)
    local w, n, a, b, poss, P, weights;
    w := CreateWordAccordingToTransitions(S, T);
    n := Length(w);
    # Slice off the last element 
    w := Subword(w, 1, n - 1);
    # Figure out what can go in the last slot
    a := Subword(w, 1, 1);
    b := Subword(w, n-1, n-1);
    poss := Filtered(S, g -> Length(a*g*b) = 3);
    # Ok, let's figure out the weighting.  If P( ) denotes the prob, then
    # P(a -> g) = (1/Sum(T[a])) T[a, g], etc. so
    # P(a -> g ->  b) = P( a ->  g ) * P(g -> b)
    P := function(x, y)
        local i_x, i_y;
        i_x := Position(S, x);
        i_y := Position(S, y);
        if Sum(T[i_x]) = 0 then 
            return 0;
        fi;
        return T[i_x][i_y]/Sum(T[i_x]);
    end;
    weights := List(poss, g -> P(a, g) * P(g, b));
    weights := weights*Lcm(List(weights, x->DenominatorRat(x)));
    if Maximum(weights) = 0 then
        weights := List(poss, g -> 1);
    fi;
    return w*RandomAccordingToDistribution(poss, weights);
end;

MakeGroupWithSameTransitions := function(G)
    local gens, rels, S, Trans, g, new_rels, F;
    
    # Make symmetric generating set
    gens := FreeGeneratorsOfFpGroup(G);
    S := [];
    for g in gens do
        Add(S, g);
        Add(S, g^-1);
    od;
    
    rels := RelatorsOfFpGroup(G);
    Trans := List(rels, rel ->  RelatorTransitions(rel, S));
    new_rels := List(Trans, T -> 
                     CreateReducedWordAccordingToTransitions(S, T));
    F := FreeGroupOfFpGroup(G);
    return F/new_rels;
end;

#------------Making groups with uniform transitions------------------

CreateUniformWord :=  function(S,len)
    local w, i, x;
    w := Random(S);
    for i in [1..len-1] do
        x := Subword(w, Length(w), Length(w));
        w := w * Random( Filtered(S, s -> not IsOne(s * x)));
    od;
    return w;
end;

CreateReducedUniformWord := function(S, len)
    local w, n;
    while true do 
        w := CreateUniformWord(S, len);
        n := Length(w);
        if not IsOne(Subword(w, 1, 1)*Subword(w, n, n)) then
            return w;
        fi;
    od;
end;
    
MakeGroupWithUniformTransitions := function(G)
    local gens, rels, S, Trans, g, new_rels, F;
    
    # Make symmetric generating set
    gens := FreeGeneratorsOfFpGroup(G);
    S := [];
    for g in gens do
        Add(S, g);
        Add(S, g^-1);
    od;
    
    rels := RelatorsOfFpGroup(G);
    new_rels := List(rels, r -> CreateReducedUniformWord(S,Length(r)));
    
    F := FreeGroupOfFpGroup(G);
    return F/new_rels;
end;

#---------Making groups with deformed presentations---------------

# Given a word finds picks n letters at random, and applies 
# and replaces them with a prob that depends on the dist. of the letters 
# in the orig. word.  The resultling word is always reduced, and differs from 
# the original one in exactly n places.

PermutePartOfWord := function(S, word, n)
    local letters, weights, poss, ind, poss_repl, p, i, a, b;
    ind := function(k) return ((k - 1) mod Length(word)) + 1; end;
    
    letters := List([1..Length(word)], i -> Subword(word, i, i));
    
    poss := [];
    for i in [1..n] do
        Add(poss, Random(Filtered([1..Length(word)], x -> not x in poss)));
    od;
     
    
    for p in poss do
        a := letters[ind(p - 1)];
        b := letters[ind(p + 1)];
        poss_repl := Filtered(S, g -> Length(a*g*b) = 3 and 
                             g <> letters[p]);
        weights := List(poss_repl, g -> Number(letters, h -> g = h));
        letters[p] := RandomAccordingToDistribution(poss_repl, weights);
    od;
    return Product(letters);
end;
 
MakeGroupWithPerturbedRelators := function(G, n)
    local gens, rels, S, Trans, g, new_rels, F;
    
    # Make symmetric generating set
    gens := FreeGeneratorsOfFpGroup(G);
    S := [];
    for g in gens do
        Add(S, g);
        Add(S, g^-1);
    od;
    
    rels := RelatorsOfFpGroup(G);
    new_rels := List(rels, r -> PermutePartOfWord(S,r,n));
    
    F := FreeGroupOfFpGroup(G);
    return F/new_rels;
end;

MakeGroupWithFivePerturb := function(G)
    return MakeGroupWithPerturbedRelators(G, 5);
end;

MakeGroupWithTwoPerturb := function(G)
    return MakeGroupWithPerturbedRelators(G, 2);
end;


MakeGroupWithOnePerturb := function(G)
    return MakeGroupWithPerturbedRelators(G, 1);
end;


#----------------------------------------------------------------
#
#  Create manifolds 
#
#----------------------------------------------------------------

MakeFakeManifold := function(M, new_group_function, name_prefix)
    local new_G, new_name, N;
    new_name := Concatenation(name_prefix, M.Name);
    new_G := new_group_function(M.Group());
    N :=CreateManifoldFromGroup(new_name, new_G);
    N.Quotients := [];
    return N;
end;    

# on 2001/7/15 created: st_simple1000 and u_simple1000
# using the initial version of this file.
# 
# STM := List(SM, M -> MakeFakeManifold(M, MakeGroupWithSameTransitions,"st_"));;   
# UM := List(SM, M -> MakeFakeManifold(M, MakeGroupWithUniformTransitions,"u_"));;   

#----------------------------------------
#
# Semi-useless testing functions
# 
#---------------------------------------

count_sign_changes := function(w)
    local e, EE, i, c;
    EE := List( [1..NumberSyllables(w)], i -> ExponentSyllable(w, i));
    c := 0;
    for i in [1..Length(EE)-1] do
        if EE[i] * EE[i+1] < 0 then
            c := c + 1;
        fi;
    od;
    if EE[1]*EE[Length(EE)] < 0 then c := c + 1; fi;
    return c;
end;

exp_sum := function(w)
    local e, EE, i, c;
    EE := List( [1..NumberSyllables(w)], i -> ExponentSyllable(w, i));
    return Sum(EE);
end;
                 
reduce_word := function(w)
    local i, W;
    W := w;
    for i in [1..Length(w)] do
        W := Subword(W, 2, Length(W)) * Subword(W, 1, 1);
    od;
    return W;
end;


#test := function(n)
#    local c, N, i, w;
#    c := 0;
#    N := Sum(Sum(T));
#    for i in [1..n] do
#        w := CreateWordAccordingToTransitions(S,T);
#        if IsOne(Subword(w, 1, 1)*Subword(w, N, N)) then
#           c := c + 1;
#        fi;
#    od;
#    return c;
#end;

#test3 := function(n)
#    local c, L, i, w, U;
#    L := List( [1..n], x -> Length(reduce_word(CreateWordAccordingToTransitions(S,T))));
#    U := Unique(L);
#    Sort(U);
#    return List(U, x -> [x, Number(L, y -> y = x)]);
#end;


#test2 := function(n, f)
#    local L, U;
#    L := List([1..n], x -> count_sign_changes(f(S,T)));
#    U := Unique(L);
#    Sort(U);
#    return List(U, x -> [x, Number(L, y -> y = x)]);
#end;

#  Stuff for testing Joe's ideas about Heegaared splittings.
#
# Below code for examining the Heegaard splitting of the cover manifold.
# G a group, f: G -> Q (finite), R a relator of G.
# Returns a list of one handles [ gen of G, starting 0-handle ]
#

LiftRelatorToCover := function(G, f, R)
    return List([1..Length(R)], i -> [Subword(R, i, i), Image(f, ElementOfFpGroup( FamilyObj(G.1), Subword(R, 1, i - 1)))]);
end;    

# The below is similar, except that the resulting list only has positively
# oriented 1-handles

AltLiftRelatorToCover := function(G, f, R)
    local F, gens;
    gens := FreeGeneratorsOfFpGroup(G);
    F := function( data )
        local g, S;
        g := data[1];
        S := data[2];
        if g in gens then
            return data;
        else
            return [g^-1, S*Image(f, ElementOfFpGroup(FamilyObj(G.1),g))];
        fi;
    end;
    return List(LiftRelatorToCover(G,f,R), F);
end;

RandomHandles := function(G, f, numToFill)
    local Lifts, Hit, L, Q, i,q;
    Lifts := List(RelatorsOfFpGroup(G), R -> AltLiftRelatorToCover(G, f, R));
    Q := Image(f);
    Hit := [];
    
    i := 0;
    while Length(Hit) < numToFill do
        q := Random(Q);
        L := Random(Lifts);
        L := List(L, dat -> [ dat[1], q*dat[2] ]);
        Hit := Union(Hit, L);
        i := i + 1;
    od;
    return i;
end;

TestRandomHandleFilling := function( Manifolds, GroupIdent, NumToTest, trials)
    local NumTested, M , G, quo_recs, quo_rec, Q, f, handles;
    NumTested := 0;
    for M in Manifolds do 
        G := M.Group();
        if hasBergeForm(G) then
            quo_recs := Filtered(M.Quotients, quo -> quo.Ident = GroupIdent);
            if Length(quo_recs) > 0 then 
                quo_rec := Random(quo_recs);
                Q := quo_rec.Image;
                f := GroupHomomorphismByImages(G, Q, 
                             GeneratorsOfGroup(G), GeneratorsOfGroup(Q));
                handles := Minimum(List([1..trials],
                                   i -> RandomHandles( G, f, (3/4)*Order(Q) )));
                Print(M.Name, "  ", List(RelatorsOfFpGroup(G), Length), "  ",
                      Order(Q), "  ", handles, "\n");
                NumTested := NumTested + 1;
            fi;
            if NumTested >= NumToTest then
                return;
            fi;
        fi;
    od;
end;

