#  $Id: rigour.gap,v 1.17 2002/06/26 20:24:25 nathand Exp $
#
#  This file is concerned with rigerously checking all of the 
#  manifolds are virtually Haken
# 
#   Here is a list of manifolds for which we had false positives
#       v3285(-4,1)  [ "Abelian" ]  cover of degree 3*43
#        s926(-4,3)  [ "LowIndex", 6, 360 ]   A_6 cover
#        s926(1,3)   [ "PSL2", 32 ]



#  The first part of the file just consists of a better, more consistant
#  set of subgroups searching and homology computation functions than
#  search.gap.  The key to making everything work smoothly is the 
#  following global:

# The global CheckHomologyGroupsSeen contains lists of pairs
#
#   [finite group, subgroups of finite group]
#
# that have already been encountered.  By convention,
#
#              (subgroups of finite group)
#
# does not consist of all the subgroups, but contains enought so that if
# G is the group, for every irreducible rep of G there is an H in 
# (subgroup list) so that the action of G on C[G/H] contains the given rep.
# The data in CheckHomologyGroupsSeen is computed via NChooseSubgroups
# in findchars.gap.

# We don't want to wipe out this knowledge when we reload "mfld_base.gap"

if not IsBound(CheckHomologyGroupsSeen) then
    CheckHomologyGroupsSeen := [];
fi;

Read("findchars.gap");

#
# FindGoodSubgroupIfAny(M, G, Q, f)
#
# given a finitely generated group G which is the fundamental group of
# a manifold M and homomorphism f:G -> Q searches for a U in Q so that
# f^-1(U) has infinite abelianization (mod a big prime).  If it finds
# such a U, it sets M.GoodCoverImage, M.GoodCoverImageU, and M.Degree
# and returns true.  Otherwise returns false

FindGoodSubgroupIfAny := function(M, G, Q, f)
    local found, dat, QQ, C, U, rank;
    
    # Check to see if we've encountered this group before.
    # If not, compute the necessary subgroups.
        
    found := false;
    for dat in CheckHomologyGroupsSeen do
        QQ := dat[1];
        if QQ = Q then
            C := dat[2];
            found := true;
            break;
        fi;
    od;
            
    if not found then
        C := NChooseSubgroups(Q);
        Add(CheckHomologyGroupsSeen, [Q, C]);
    fi;
    
    # Having found the subgroups
    
    for U in C do 
        rank := AbelianRankOfSubgroup(G, PreImage(f, U));
        if rank > 0 then
            M.GoodCoverImage := List(GeneratorsOfGroup(G), g -> Image(f, g));
            M.GoodCoverImageU := GeneratorsOfGroup(U);
            M.Degree := Index(Q, U);
            M.KnownWeakPosBettiCover := true;
            Print("  Success at index ", M.Degree, "\n");
            return true;
        fi;
    od;
    return false;
end;

# Given a manifold with KnownWeakPosBettiCover := true and a stored
# cover in GoodCoverImage and GoodCoverImageU, use checker to
# rigorously compute the rank of that cover.  

CheckHomologyWithBlah := function(M, checker, reason)
    local  C, G, Q, f, R, U, H, dat, found;
    
    if not M.KnownPosBettiCover and M.KnownWeakPosBettiCover and 
       Length(M.GoodCoverImage) > 1 then
        Print(M.Name, "  ", M.GoodCover);
        G := M.Group();
        Q := Group(M.GoodCoverImage);
        f := GroupHomomorphismByImages(G, Q, 
                     GeneratorsOfGroup(G), GeneratorsOfGroup(Q));
        
        if not IsBound(M.GoodCoverImageU) then
            if not FindGoodSubgroupIfAny(M, G, Q, f) then
                Print("\n ERROR! no good subgroup\n");
            fi;
            
        fi;
        if Length(M.GoodCoverImageU) > 0 then
            U := Group(M.GoodCoverImageU);
        else
            U := Group( () );
        fi;
        
            
        Print("  ", M.Degree, "\n");
        #return;
        #return AbelianRankOfSubgroup(G, PreImage(f,U));;
        #return RelatorMatrixAbelianizedSubgroupRrs(G, PreImage(f,U));
        R := checker(G, PreImage(f,U));
        if R > 0 then
            M.KnownPosBettiCover := true;
            M.Reason := reason;
            M.Rank := R;
            Print("   ", M.Degree, "  ", R, "  ", Order(Q),  "\n");
        else
            Print("   FAILURE: no Q homology\n");
        fi;
    fi;
end;

CheckHomologyWithAbelianInvariants := function(M)
    CheckHomologyWithBlah(M, AbelianInvariantRankOfSubgroup,
            "AbelianInvariants");
end;

CheckHomologyWithRationalRecon := function(M)
    CheckHomologyWithBlah(M, TrueAbelianRankOfSubgroup, 
            "RationalReconstruction");
end; 

  
# The next function makes use of the groups in CheckHomologyGroupsSeen 
# to speed computation.  For large n, the killer is computing all the 
# subgroups of S_n and A_n.  NChooseSubgroups from findchars.gap makes some
# approximation to these subgroups when n > 9.  Returns true if it finds
# a good subgroup.  Note:  Users should to use the function
# SearchLowIndexSubgroups instead.

CheckLowIndexSubgroups := function(M, min_index, max_index)
    local f, fs, G, i, lilist, H, Q, data;
    
    Print(M.Name, "\n");
    G := M.Group();
    lilist := LowIndexSubgroupsFpGroup(G, TrivialSubgroup(G), max_index);
    lilist := Filtered(lilist, H -> (Index(G, H) >= min_index and 
                      Index(G,H) <= max_index));
    Sort( lilist, function(g,h) return Index(G,g) <= Index(G,h); end);
    
    # Now take the smallest normal subgroup containing each of lilist
    # and sort them via the new index
            
    fs := List(lilist, H -> FactorCosetAction(G, H));
    data := List(fs, f -> [f, Image(f)]);
    Sort(data, function(d1, d2) return Order(d1[2]) <= Order(d2[2]); end);
    Print("   ", List(lilist, H -> Index(G, H)), "\n");
    Print("   ", List(data, dat -> Order(dat[2])), "\n");
    for i in [1..Length(data)] do
        f := data[i][1];
        Q := data[i][2];
        if FindGoodSubgroupIfAny(M, G, Q, f) then
            M.GoodCover := ["LowIndex", NrMovedPoints(Q)];
            return true;
        fi;
    od; 
    return false;
end;

# Note:  Users should to use the Search* functions instead

CheckSubgroupsWithGivenQuotient := function(M, Q)
    local f, G;
    Print(M.Name, "  looking for quotient of order ", Order(Q), "\n");
    G := M.Group();
    for f in GQuotients(G, Q) do
        if FindGoodSubgroupIfAny(M, G, Q, f) then
            return true;
        fi;
    od;
    return false;
end;

#--------------------------------------------------------------------------
#
#  New searching functions superceeding those in search.gap
#
#--------------------------------------------------------------------------

SearchForAbelianCover := function(M)
    local G, f, N;
    if ManifoldAlmostDone(M) then
        return;
    fi;
    Print(M.Name, "\n");
    G := M.Group();
    N := DerivedSubgroup(G);
    f := FactorCosetAction(G, N);
    if FindGoodSubgroupIfAny(M, G, Image(f), f) then
        M.GoodCover := ["Abelian"];
    fi;
end;

# Follows convention that n <= 6 means search all 
# subgroups of index less than n and n > 6 implies
# the we search all subgroups of index = n.

SearchForLowIndexCover := function(M, n)
    local success, min_n, max_n;
    if ManifoldAlmostDone(M) then
        return;
    fi;
    if n <= 6 then
        min_n := 1;
    else
        min_n := n;
    fi;
    max_n := n;
    CheckLowIndexSubgroups(M, min_n, max_n);
end;

SearchForPerfectCover := function(M, ident)
    if ManifoldAlmostDone(M) then
        return;
    fi;
    if CheckSubgroupsWithGivenQuotient(M, PerfectGroup(IsPermGroup, ident)) then
        M.GoodCover := ["Perfect", ident, PerfectGroupName(ident)];
    fi;
end;

SearchForPSL2Cover := function(M, p)
    if ManifoldAlmostDone(M) then
        return;
    fi;
    if CheckSubgroupsWithGivenQuotient(M, PSL(2,p)) then
        M.GoodCover := ["PSL2", p];
    fi;
end;

# Below is a simpler version of the tensor product of representation stuff in
# search.gap.  Note that the alternate version is probably better.

SearchForProductCover := function(M, max_index)
    local f, fs, G, i, j, lilist, H, Q, data, f1, Q1, f2, Q2, QP, e1, e2,
          prod_image;
    if ManifoldAlmostDone(M) then
        return;
    fi;
    
    Print(M.Name, "\n");
    G := M.Group();
    lilist := LowIndexSubgroupsFpGroup(G, TrivialSubgroup(G), max_index);
    lilist := Filtered(lilist, H -> (Index(G, H) > 1 and 
                      Index(G,H) <= max_index));
    Sort( lilist, function(g,h) return Index(G,g) <= Index(G,h); end);
    
    # Now take the smallest normal subgroup containing each of lilist
    # and sort them via the new index
            
    fs := List(lilist, H -> FactorCosetAction(G, H));
    data := List(fs, f -> [f, Image(f)]);
    Sort(data, function(d1, d2) return Order(d1[2]) <= Order(d2[2]); end);
    Print("   ", List(lilist, H -> Index(G, H)), "\n");
    Print("   ", List(data, dat -> Order(dat[2])), "\n");
    for i in [1..Length(data)] do
        for j in [1..(i-1)] do
            f1 := data[i][1];
            Q1 := data[i][2];
            f2 := data[j][1];
            Q2 := data[j][2];
            QP := DirectProduct(Q1, Q2);
            e1 := Embedding(QP,1);
            e2 := Embedding(QP,2);
            
            prod_image := List(GeneratorsOfGroup(G), 
                               g -> Image(e1, Image(f1, g)) *
                               Image(e2, Image(f2, g)));
            Q := Group(prod_image);
            f := GroupHomomorphismByImages(G, Q, GeneratorsOfGroup(G),
                         prod_image);
            Print("    trying: ", [Order(Q1), Order(Q2), Order(Q)] , "\n");
            if FindGoodSubgroupIfAny(M, G, Q, f) then
                M.GoodCover := ["IntersectionOfLowIndex", 
                                [NrMovedPoints(Q1), NrMovedPoints(Q2)] ];
                return true;
            fi;
        od; 
    od;
    return false;
end;


# Alternate version which useses less memory so can handle
# "v3372(3,2)".  Though this is the "Alt" version, quite possibly
# it is better that the standard one.  

SearchForProductCoverAlt := function(M, max_index, max_degree_of_cover)
    local f, fs, G, i, j, lilist, H, Q, data, f1, Q1, f2, Q2, QP, e1, e2,
          prod_image, U1s, U2s, U, U1, U2, Us;
    if ManifoldAlmostDone(M) then
        return;
    fi;
    
    Print(M.Name, "\n");
    G := M.Group();
    lilist := LowIndexSubgroupsFpGroup(G, TrivialSubgroup(G), max_index);
    lilist := Filtered(lilist, H -> (Index(G, H) > 1 and 
                      Index(G,H) <= max_index));
    Sort( lilist, function(g,h) return Index(G,g) <= Index(G,h); end);
    
    # Now take the smallest normal subgroup containing each of lilist
    # and sort them via the new index
            
    fs := List(lilist, H -> FactorCosetAction(G, H));
    data := List(fs, f -> [f, Image(f)]);
    Sort(data, function(d1, d2) return Order(d1[2]) <= Order(d2[2]); end);
    Print("   ", List(lilist, H -> Index(G, H)), "\n");
    Print("   ", List(data, dat -> Order(dat[2])), "\n");
    for i in [1..Length(data)] do
        for j in [1..(i-1)] do
            f1 := data[i][1];
            Q1 := data[i][2];
            f2 := data[j][1];
            Q2 := data[j][2];
            QP := DirectProduct(Q1, Q2);
            e1 := Embedding(QP,1);
            e2 := Embedding(QP,2);
            
            prod_image := List(GeneratorsOfGroup(G), 
                               g -> Image(e1, Image(f1, g)) *
                               Image(e2, Image(f2, g)));
            Q := Group(prod_image);
            f := GroupHomomorphismByImages(G, Q, GeneratorsOfGroup(G),
                         prod_image);
            Print("    trying: ", [Order(Q1), Order(Q2), Order(Q)] , "\n");
            
            # compute possible U 
            
            U1s := NChooseSubgroups(Q1);
            U2s := NChooseSubgroups(Q2);
            Us := [];
            for U1 in U1s do
                for U2 in U2s do
                    U := Group(Concatenation(
                            List(GeneratorsOfGroup(U1), u1 -> Image(e1, u1)),
                                 List(GeneratorsOfGroup(U2), u2 -> Image(e2, u2))));
                    U := Intersection(U, Q);
                    if Index(Q, U) <= max_degree_of_cover then
                        Add(Us, U);
                    fi;
                od;
            od;
            Sort(Us, function(g,h) return Index(Q,g) <= Index(Q,h); end);
            Us := Unique(Us);
            Add(CheckHomologyGroupsSeen, [Q, Us]);
            
            Print("\nAbout to look at subgroups of index:\n");
            Print("    ", List(Us, u -> Index(Q,u)), "\n\n");
            if FindGoodSubgroupIfAny(M, G, Q, f) then
                M.GoodCover := ["IntersectionOfLowIndex", 
                                [NrMovedPoints(Q1), NrMovedPoints(Q2)] ];
                return true;
            fi;
        od; 
    od;
    return false;
end;


# Hack for attacking the last manifold.  

AttackLast := function(M, min_degree_of_cover, max_degree_of_cover)
    local f, fs, G, i, j, lilist, H, Q, data, f1, Q1, f2, Q2, QP, e1, e2,
          prod_image, U1s, U2s, U, U1, U2, Us;
    if ManifoldAlmostDone(M) then
        return;
    fi;
    
    Print(M.Name, "\n");
    G := M.Group();
    lilist := List(M.SimpleGroupQuotients, x -> x.Image);
            
    fs := List(lilist, H -> GroupHomomorphismByImages(G, H, GeneratorsOfGroup(G), GeneratorsOfGroup(H)));
    data := List(fs, f -> [f, Image(f)]);
    Sort(data, function(d1, d2) return Order(d1[2]) <= Order(d2[2]); end);
    Print("   ", List(lilist, H -> Order(H)), "\n");
    Print("   ", List(data, dat -> Order(dat[2])), "\n");
    for i in [1..Length(data)] do
        for j in [1..(i-1)] do
            f1 := data[i][1];
            Q1 := data[i][2];
            f2 := data[j][1];
            Q2 := data[j][2];
            QP := DirectProduct(Q1, Q2);
            e1 := Embedding(QP,1);
            e2 := Embedding(QP,2);
            
            prod_image := List(GeneratorsOfGroup(G), 
                               g -> Image(e1, Image(f1, g)) *
                               Image(e2, Image(f2, g)));
            Q := Group(prod_image);
            f := GroupHomomorphismByImages(G, Q, GeneratorsOfGroup(G),
                         prod_image);
            Print("    trying: ", [Order(Q1), Order(Q2), Order(Q)] , "\n");
            
            # compute possible U 
            
            U1s := NChooseSubgroups(Q1);
            U2s := NChooseSubgroups(Q2);
            Us := [];
            for U1 in U1s do
                for U2 in U2s do
                    U := Group(Concatenation(
                            List(GeneratorsOfGroup(U1), u1 -> Image(e1, u1)),
                                 List(GeneratorsOfGroup(U2), u2 -> Image(e2, u2))));
                    U := Intersection(U, Q);
                    if Index(Q, U) <= max_degree_of_cover and
                       Index(Q, U) > min_degree_of_cover then
                        Add(Us, U);
                    fi;
                od;
            od;
            Sort(Us, function(g,h) return Index(Q,g) <= Index(Q,h); end);
            Us := Unique(Us);
            Add(CheckHomologyGroupsSeen, [Q, Us]);
            
            Print("\nAbout to look at subgroups of index:\n");
            Print("    ", List(Us, u -> Index(Q,u)), "\n\n");
            if FindGoodSubgroupIfAny(M, G, Q, f) then
                M.GoodCover := ["IntersectionOfLowIndex", 
                                [NrMovedPoints(Q1), NrMovedPoints(Q2)] ];
                return true;
            fi;
        od; 
    od;
    return false;
end;


#-------------------------------------------------------------------------
#
#  This section gives the code used to extract the data from 
#  "mflds-3-21.gap" which was our first check of all the manifolds
#  It is of no interest to others.  
#  
#-------------------------------------------------------------------------


TransferAbelianSubgroups := function(CM, FM)
    local M, f, G, N, i;
    for i in [1..Length(CM)] do
        M := CM[i];
        if M.GoodCover = "Abelian Cover" then
            G := M.Group();
            N := DerivedSubgroup(G);
            f := FactorCosetAction(G, N);

            FM[i].KnownWeakPosBettiCover := true;
            FM[i].GoodCover := ["Abelian"];
            FM[i].GoodCoverImage := List(GeneratorsOfGroup(G), g -> Image(f, g));
        fi;
    od;
end;

TrivialCovers := function(FM)
    local R, M;
    for M in FM do
        R := Number(M.Homology, x -> x = 0);
        if R > 0 then
            Assert(-1, M.KnownWeakPosBettiCover = false);
            M.KnownWeakPosBettiCover := true;
            M.KnownPosBettiCover := true;
            M.GoodCover := ["Trivial"];
            M.Reason := "AbelianInvariants";
            M.Degree := 1;
            M.Rank := R;
        fi;
    od;
end;

TransferLowIndex := function(CM, FM)
    local M, f, G, N, index, i, type;
    for i in [1..Length(CM)] do
        M := CM[i];
        N := FM[i];
        if M.GoodCover <> false  and not N.KnownPosBettiCover then 
            type := String(M.GoodCover[1]);


            if Length(M.GoodCover) >= 1 and  
               type[1] = 'L' then
                if type = "LowIndex" then
                    index := M.GoodCover;
                    N.KnownWeakPosBettiCover := true;
                    N.GoodCover := M.GoodCover;
                fi;
                if type = "Low Index" then
                    N.KnownWeakPosBettiCover := true;
                    if M.GoodCover[2] > 10 then 
                        N.GoodCover := ["LowIndex", 6, M.GoodCover[2]];
                    else
                        N.GoodCover := ["LowIndex", 6];
                    fi;
                fi;
            fi;
        fi;
    od;
end;

TransferPSLIndex := function(CM, FM)
    local M, f, G, N, index, i, type;
    for i in [1..Length(FM)] do
        N := FM[i];
        M := GetManifold(CM, N.Name);
        if M <> fail and M.GoodCover <> false and not N.KnownPosBettiCover then 
            type := String(M.GoodCover[1]);
            

            if Length(M.GoodCover) >= 1 and  
               (type = "PSL" or type = "PSL2") then
                Print(N.Name, "  ", M.GoodCover[2], "\n");
                N.KnownWeakPosBettiCover := true;
                N.GoodCover := ["PSL2", M.GoodCover[2]];
            fi;
            if M.GoodCover = "PSL(2,2)" then
                N.KnownWeakPosBettiCover := true;
                N.GoodCover := ["LowIndex", 6];
            fi;
            if M.GoodCover = "PSL(2,7)" then
                N.KnownWeakPosBettiCover := true;
                N.GoodCover := ["PSL2", 7];
            fi;
        fi;
    od;
end;

TransferPerfectAndSimple := function(CM, FM)
    local M, f, G, N, index, i, type, ident;
    for i in [1..Length(CM)] do
        M := CM[i];
        N := FM[i];
        if M.GoodCover <> false  and not N.KnownWeakPosBettiCover then 
            if Length(M.GoodCover) >= 1 and  
               M.GoodCover[1] = "Perfect" then
                N.KnownWeakPosBettiCover := true;
                ident := M.GoodCover[2];
                N.GoodCover := ["Perfect", ident, PerfectGroupName(ident)];
            fi;
            if Length(M.GoodCover) >= 1 and  
               M.GoodCover[1] = "Simple" then
                N.KnownWeakPosBettiCover := true;
                ident := ConvertSimpleIdentToPerfect(M.GoodCover[2]);
                N.GoodCover := ["Perfect", ident, PerfectGroupName(ident)];
            fi;

        fi;
    od;
end;

TransferTensor := function(CM, FM)
    local M, f, G, N, index, i, type, ident;
    for i in [1..Length(CM)] do
        M := CM[i];
        N := FM[i];
        if M.GoodCover <> false  and not N.KnownWeakPosBettiCover then 
            if Length(M.GoodCover) >= 1 and  
               M.GoodCover[1] = "TensorOfReps" then
                N.GoodCover := M.GoodCover;
            fi;
        fi;
    od;
end;

TransferPGroup := function(CM, FM)
    local M, f, G, N, index, i, type, ident;
    for i in [1..Length(CM)] do
        M := CM[i];
        N := FM[i];
        if M.GoodCover <> false  and not N.KnownWeakPosBettiCover then 
            if M.Reason = "p-group" then
                N.KnownWeakPosBettiCover := true;
                N.GoodCover := ["p-group", M.GoodCover];
            fi;
        fi;
    od;
end;


ConstructDoubleImage := function(G)
    local gens;
    gens := GeneratorsOfGroup(G);
    return [gens[2]*gens[1], gens[3]*gens[1]];
end;
    
TransferLowIndexDouble := function(CM, FM, IM)
    local M, N, K,n,L;
    for N in ManifoldsWithoutWeakPosBettiCovers(FM) do
        M := GetManifold(CM, N.Name);
        if M <> fail and M.GoodCover <> false and M.GoodCover[1] = "LowIndexCoverDouble" then
            n := M.GoodCover[2];
            K := GetManifold(IM, N.Name);
            L := Filtered(K.DImages, r -> r.PermDegree = n);
            if Length(L) > 1 then
                Print("Poss. prob\n");
            fi;
            
            if Length(L) > 0 then 
                N.KnownWeakPosBettiCover := true;
                N.GoodCover := ["LowIndex", n];
                N.GoodCoverImage := ConstructDoubleImage(L[1].Group);
                Print(N.Name,"Ok \n");
            else
                Print(N.Name,"fail \n");
            fi;
                  
        fi;
    od;
end;
    
    
TransferPSLFromImageData := function(FM, IM)
    local M, N, K,n,L, quo_rec, G, f, Q;
    for M in Filtered(ManifoldsNotAlmostDone(FM), m-> m.GoodCover <> false) do
        if M.GoodCover[1] = "PSL2" then
            n := M.GoodCover[2];
            N := GetManifold(IM, M.Name);
            if n in N.Scans.PSL then
                for quo_rec in N.Images do
                    if quo_rec.GroupType = ["L", [2,n]] then
                        Print(M.Name, "\n");
                        G := M.Group();
                        Q := quo_rec.Group;
                        f := GroupHomomorphismByImages(G, Q, 
                                     GeneratorsOfGroup(G), GeneratorsOfGroup(Q));
                        if FindGoodSubgroupIfAny(M, G, Q, f) then
                            break;
                        fi;
                    fi;
                od;
            fi;
        fi;
    od;
end;
            

UntransferedTypes := function(CM, FM)
    local M, L;
    L := [];
    for M in Filtered(FM, M->M.GoodCover = false) do
        Add(L, GetManifold(CM, M.Name).GoodCover);
    od;
    Sort(L);
    return L;
end;

TransferGoodCoverHom := function(CM, FM)
    local M, N, Q;
    for N in ManifoldsNotAlmostDone(FM) do
        M := GetManifold(CM, N.Name);
        if M <> fail and IsBound(M.GoodCoverHom) then
            Assert(-1, N.GoodCover[1] = "Perfect");
            Assert(-1, N.GoodCover[2][1] = Order(M.GoodCoverHom));
            N.GoodCoverImage := GeneratorsOfGroup(M.GoodCoverHom);
            CheckHomologyWithAbelianInvariants(N);
        fi;
    od;
end;
    
            

ConvertFakePSL2sToSimple := function(M)
    if M.GoodCover <> false and Length(M.GoodCover) > 0 then
        if M.GoodCover[1] = "PSL2" and not IsInt(M.GoodCover[2]) then
            Print(M.Name, "  ", M.GoodCover, "\n");
            if M.GoodCover[2] = [3,3] then
                M.GoodCover := ["Perfect", [ 5616, 1 ], "L3(3)"];
            fi;
            if M.GoodCover[2] = [3,4] then
                M.GoodCover := ["Perfect", [ 20160, 5 ], "L3(4)"];
            fi;
        fi;
    fi;
end;

            
            
            
FindCorrectPGroupSubgroup := function(M)
    local Q, G, f,g, h, p,i, A;
    if not M.KnownPosBettiCover and M.KnownWeakPosBettiCover and
       Length(M.GoodCoverImage) = 0 and M.GoodCover[1] = "p-group" then
        p := M.GoodCover[2][1];
        i := M.GoodCover[2][2];
        G := M.Group();
        g := EpimorphismPGroup(G, p, i);
        f := IsomorphismPermGroup(Image(g));
        h := CompositionMapping(f,g);
        g := SmallerDegreePermutationRepresentation(Image(h));
        f := CompositionMapping(g, h);
        FindGoodSubgroupIfAny(M, G, Image(f), f);
    fi;
end;



FindCorrectPSLSubgroups := function(M, p)
    if not M.KnownPosBettiCover and M.KnownWeakPosBettiCover and
       Length(M.GoodCoverImage) = 0 and M.GoodCover[1] = "PSL2" then
        if M.GoodCover[2] = p then
            CheckSubgroupsWithGivenQuotient(M, PSL(2,p));
        fi;
    fi;
end;

FindCorrectPerfectSubgroups := function(M)
    if not M.KnownPosBettiCover and M.KnownWeakPosBettiCover and
       Length(M.GoodCoverImage) = 0 and M.GoodCover[1] = "Perfect" then
        CheckSubgroupsWithGivenQuotient(M, PerfectGroup(IsPermGroup, M.GoodCover[2]));
    fi;
end;


FindCorrectLowIndexSubgroup := function(M, n)
    local index;
    if not M.KnownPosBettiCover and M.KnownWeakPosBettiCover and
       M.GoodCover[1] = "LowIndex" and Length(M.GoodCoverImage) = 0 then
        index := M.GoodCover[2];
        if index = n then
            if n > 6 then
                CheckLowIndexSubgroups(M, n, n);
            else
                CheckLowIndexSubgroups(M, 1, n);
            fi;
        fi;
    fi;
end;




        
        
